//! Code which enables sending messages between processes in the blocktree system.
use btlib::{
    Result,
    crypto::rand_array,
    BlockPath, Writecap,
};
use std::{
    path::PathBuf,
    io::{Read, Write},
    os::unix::net::{UnixStream, UnixListener}, 
};
use btserde::{read_from, write_to};
use serde::{Deserialize, Serialize};
use zerocopy::FromBytes;

pub use private::*;

mod private {
    use std::io;

    use super::*;

    /// The directory where sockets are created.
    fn socket_dir() -> PathBuf {
        let home_dir: PathBuf = std::env::var("HOME").unwrap().into();
        home_dir.join(".btmsg")
    }

    pub fn socket_path(block_path: &BlockPath) -> PathBuf {
        socket_dir().join(block_path.to_string())
    }

    #[derive(PartialEq, Eq, Serialize, Deserialize)]
    pub enum MsgError {}

    #[derive(Deserialize)]
    pub enum MsgBodyOwned {
        Success,
        Fail(MsgError),
        Ping,
        Hello(Writecap),
        Read { offset: u64, size: u64, },
        Write { offset: u64, buf: Vec<u8> },
        Custom(Vec<u8>),
    }

    #[derive(Serialize)]
    pub enum MsgBodyRef<'a> {
        Success,
        Fail(&'a MsgError),
        Ping,
        Hello(&'a Writecap),
        Read { offset: u64, buf: &'a [u8] },
        Write { offset: u64, buf: &'a mut [u8] },
        Custom(&'a [u8]),
    }

    pub trait Msg {
        type Body;
        fn from(&self) -> &BlockPath;
        fn to(&self) -> &BlockPath;
        fn id(&self) -> u128;
        fn body(&self) -> &Self::Body;
    }

    #[derive(Deserialize)]
    pub struct MsgOwned {
        from: BlockPath,
        to: BlockPath,
        id: u128,
        body: MsgBodyOwned,
    }

    impl Msg for MsgOwned {
        type Body = MsgBodyOwned;

        fn from(&self) -> &BlockPath {
            &self.from
        }

        fn to(&self) -> &BlockPath {
            &self.to
        }

        fn id(&self) -> u128 {
            self.id
        }

        fn body(&self) -> &Self::Body {
            &self.body
        }
    }

    #[derive(Serialize)]
    pub struct MsgRef<'a> {
        from: &'a BlockPath,
        to: &'a BlockPath,
        id: u128,
        body: MsgBodyRef<'a>,
    }

    impl<'a> Msg for MsgRef<'a> {
        type Body = MsgBodyRef<'a>;

        fn to(&self) -> &BlockPath {
            &self.to
        }

        fn from(&self) -> &BlockPath {
            &self.from
        }

        fn id(&self) -> u128 {
            self.id
        }

        fn body(&self) -> &Self::Body {
            &self.body
        }
    }

    #[derive(Deserialize)]
    pub enum VerMsg {
        V0(MsgOwned),
    }

    #[derive(Serialize)]
    pub enum VerMsgRef<'a> {
        V0(MsgRef<'a>),
    }

    pub trait Sender: Send {
        fn send<'a>(&mut self, msg: MsgRef<'a>) -> Result<()>;
        fn path(&self) -> &BlockPath;

        /// Generates and returns a new message ID.
        fn gen_id(&mut self) -> Result<u128> {
            const LEN: usize = std::mem::size_of::<u128>();
            let bytes = rand_array::<LEN>()?;
            let option = u128::read_from(bytes.as_slice());
            // Safety: because LEN == size_of::<u128>(), read_from should have returned Some.
            Ok(option.unwrap())
        }

        fn send_new<'a>(&mut self, to: &BlockPath, from: &BlockPath, body: MsgBodyRef<'a>) -> Result<()> {
            let id = self.gen_id()?;
            let msg = MsgRef {
                to,
                from,
                id,
                body,
            };
            self.send(msg)
        }
    }

    pub trait Receiver {
        fn receive(&mut self) -> Result<MsgOwned>;
        fn path(&self) -> &BlockPath;
    }

    pub trait Channel: Sized + Receiver {
        type Sender: Sender;

        fn new(receiver_path: BlockPath) -> Result<Self>;
        fn connect_to(receiver_path: BlockPath) -> Result<Self::Sender>;
    }

    pub struct WriteSender<W> {
        write: W,
        path: BlockPath,
    }

    impl<W> WriteSender<W> {
        pub fn new(write: W, path: BlockPath) -> Self {
            Self { write, path }
        }
    }

    impl<W: Write + Send> Sender for WriteSender<W> {
        fn send<'a>(&mut self, msg: MsgRef<'a>) -> Result<()> {
            Ok(write_to(&VerMsgRef::V0(msg), &mut self.write)?)
        }

        fn path(&self) -> &BlockPath {
            &self.path
        }
    }

    pub struct ReadReceiver<R> {
        read: R,
        path: BlockPath,
    }

    impl<R> ReadReceiver<R> {
        pub fn new(read: R, path: BlockPath) -> Self {
            Self { read, path }
        }
    }

    impl<R: Read> Receiver for ReadReceiver<R> {
        fn receive(&mut self) -> Result<MsgOwned> {
            let VerMsg::V0(msg) = read_from(&mut self.read)?;
            Ok(msg)
        }

        fn path(&self) -> &BlockPath {
            &self.path
        }
    }

    pub struct UnixChannel {
        receiver_path: BlockPath,
        listener: UnixListener,
        streams: Vec<ReadReceiver<UnixStream>>,
        next: usize,
    }

    impl Receiver for UnixChannel {
        fn path(&self) -> &BlockPath {
            &self.receiver_path
        }

        fn receive(&mut self) -> Result<MsgOwned> {
            // Note that the listener is put in non-blocking mode when it is created.
            match self.listener.accept() {
                Ok((socket, ..)) => {
                    self.streams.push(ReadReceiver::new(socket, self.receiver_path.clone()));
                }
                Err(err) => {
                    // If the error is anything other than `WouldBlock`, then it is unexpected.
                    if io::ErrorKind::WouldBlock != err.kind() {
                        return Err(err.into());
                    }
                }
            }
            let receiver = self.streams.get_mut(self.next).unwrap();
            self.next += 1;
            receiver.receive()
        }
    }

    impl Channel for UnixChannel {
        type Sender = WriteSender<UnixStream>;

        fn new(receiver_path: BlockPath) -> Result<Self> {
            let path = socket_path(&receiver_path);
            std::fs::create_dir_all(path.parent().unwrap())?;
            let listener = UnixListener::bind(&path)?;
            listener.set_nonblocking(true)?;
            Ok(UnixChannel { receiver_path, listener, streams: Vec::new(), next: 0 })
        }

        fn connect_to(receiver_path: BlockPath) -> Result<Self::Sender> {
            let path = socket_path(&receiver_path);
            let stream = UnixStream::connect(&path)?;
            let sender = WriteSender::new(stream, receiver_path);
            Ok(sender)
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    //#[test]
    //fn ping_pong_via_pipes() {
    //    let ping_path = test_helpers::make_path(vec!["ping"]);
    //    let (mut ping_sender, mut ping_receiver) =
    //        PipeChannel::new(ping_path.clone()).expect("failed to create a ping channel");
    //    let pong_path = test_helpers::make_path(vec!["pong"]);
    //    let (mut pong_sender, mut pong_receiver) =
    //        PipeChannel::new(pong_path.clone()).expect("failed to create a pong channel");
    //    let pong = std::thread::spawn(move || loop {
    //        let msg = pong_receiver.receive().expect("pong receive failed");
    //        assert_eq!(pong_receiver.path(), msg.to());
    //        match msg.body() {
    //            MsgBody::Ping => ping_sender
    //                .send_new(ping_path.clone(), MsgBody::Success)
    //                .expect("send to ping failed"),
    //            MsgBody::Success => return,
    //            _ => panic!("unexpected message received by pong"),
    //        }
    //    });
    //    let mut iter = 5;
    //    while iter > 0 {
    //        pong_sender
    //            .send_new(pong_path.clone(), MsgBody::Ping)
    //            .expect("send to pong failed");
    //        let msg = ping_receiver.receive().expect("ping receive failed");
    //        assert_eq!(ping_receiver.path(), msg.to());
    //        match msg.body() {
    //            MsgBody::Success => iter -= 1,
    //            _ => panic!("unexpected message received by ping"),
    //        }
    //    }
    //    pong_sender
    //        .send_new(pong_path.clone(), MsgBody::Success)
    //        .expect("send success to pong failed");
    //    pong.join().expect("join failed");
    //}
}
